/* upward_opt.c
   28-JAN-1998, David Mathog, Biology Division, Caltech

   Processes two linker .opt files "old" "new".  It produces
   from them a "final" linker .opt file which is an upwardly
   compatible version of "new" which should work with code linked
   with "old".

   This is NOT optimized for speed!!!

   invoke like:

   upward_opt old.opt new.opt final.opt major minor

   where nullfunction is the name of the function contained in 
   the new library which will be called when a program invokes
   a function which has been removed.  It should issue a warning
   and then cause the program to crash with a stack dump.

   Only these types of lines are recognized:

SYMBOL_VECTOR = (FL_INVERTED_Y=PSECT)
SYMBOL_VECTOR = (FL_INVERTED_Y=PRIVATE_DATA)
SYMBOL_VECTOR = (FL_DASHEDLINESTYLE=PROCEDURE)
SYMBOL_VECTOR = (FL_DASHEDLINESTYLE=PRIVATE_PROCEDURE)
GSMATCH=lequal,N,M

   Everything else is a "verbatim" line, and goes through unchanged.

   In this version all PRIVATE_PROCEDUREs are called MISSINGFCT

*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

FILE *fold;
FILE *fnew;
FILE *ffinal;
#define MAXINLINE 1000
char inline[MAXINLINE];
int major, minor; /* version numbers for GSMATCH*/

enum hasstate {UNKNOWN,SAME,REMOVED,NEW,VERBATIM};  /* state of each line */
enum hastype  {GSMATCH,PSECT,PROCEDURE,PRIVATE_PROCEDURE,PRIVATE_DATA,OTHER};  /* type of of each line */
typedef struct isaline ALINE;
struct isaline {             /* line node */
   ALINE *next;
   enum hasstate state;
   enum hastype type;
   char *string;
   char *comment;      /* place to save comments */
};
ALINE *head=NULL;       /* list of line nodes */
ALINE *tail=NULL;
ALINE *current=NULL;


void oops(char * string){
 (void) fprintf(stderr,"upward_opt, fatal error, %s\n",string);
 exit(EXIT_FAILURE);
}

void savecomment(void){
char *front;
  front = strchr(inline,'!');
  if(front == NULL){
    current->comment = malloc(1);
    if(current->comment == NULL)oops("could not allocate enough memory");
    *(current->comment)='\0';
  }
  else {
    current->comment = malloc(1+strlen(front));
    if(current->comment == NULL)oops("could not allocate enough memory");
    (void)strcpy(current->comment,front);
    *front = '\0';
  }  
}  

void addstring(char *back){
char *front;
  *back='\0';
  for(front=back; (*front != '(') && (front != &inline[0]); front--){
  }
  if(front != &inline[0])front++;
  current->string = malloc(1+strlen(front));
  if(current->string == NULL)oops("could not allocate enough memory");
  (void)strcpy(current->string,front);
}  


void characterize(void){
char *front;
char *back;

  /* squish the \n on the end of the line */
  front = strchr(inline,'\n');
  if(front != NULL)*front='\0';

  current = malloc(sizeof(ALINE));
  if(current == NULL)oops("could not allocate enough memory");
  current->next = NULL;
  savecomment();

  if((back=strstr(inline,"GSMATCH=lequal")) != NULL     ){
     current->type=GSMATCH;
     current->state=UNKNOWN;
     current->string = malloc(1+strlen(&inline[0]));
     if(current->string == NULL)oops("could not allocate enough memory");
     (void)strcpy(current->string,inline);
     return;
  }
  if((back=strstr(inline,"=PSECT)")) != NULL     ){
     current->type=PSECT;
     current->state=UNKNOWN;
     addstring(back);
     return;
  }
  if((back=strstr(inline,"=PRIVATE_DATA)")) != NULL){
     current->type=PRIVATE_DATA;
     current->state=UNKNOWN;
     addstring(back);
     return;
  }
  if((back=strstr(inline,"=PROCEDURE)")) != NULL){
     current->type=PROCEDURE;
     current->state=UNKNOWN;
     addstring(back);
     return;
  }
  if((back=strstr(inline,"=PRIVATE_PROCEDURE)")) != NULL){
     current->type=PRIVATE_PROCEDURE;
     current->state=UNKNOWN;
     addstring(back);
     return;
  }
  current->type=OTHER;
  current->state=VERBATIM;
  current->string = malloc(1+strlen(&inline[0]));
  if(current->string == NULL)oops("could not allocate enough memory");
  (void)strcpy(current->string,inline);
}

void readold(void){
  while (fgets(inline,MAXINLINE-1,fold) != NULL){
    characterize();
    if(strlen(current->string) + strlen(current->comment) >0){  /* remove blank lines */
      if(tail == NULL){
        head = current;
        tail = current;
      }
      else {
        tail->next=current;
        tail = current;
      }
    }
  }
  strcpy(inline,"! end of OLD linker option file");
  characterize();
  tail->next=current;
  tail = current;
}

void readnew(void){
ALINE *scanner;
int foundit;
  if(tail == NULL || head == NULL)oops("old option file was empty, something is very wrong!");
  while (fgets(inline,MAXINLINE-1,fnew) != NULL){
    characterize();
    if(strlen(current->string) >0){  /* comments are not preserved */
      for(foundit=0, scanner=head; scanner != NULL ; scanner=scanner->next){
        if(strcmp(scanner->string,current->string) == 0){  /* match */
          foundit=1;
	  if(scanner->type != current->type){ /* big error, name changed type*/
            (void) fprintf(stderr,"upward_opt, fatal error, mismatched types for %s\n",scanner->string);
          }
          else {
            scanner->state = SAME;
            free(scanner->comment);
            scanner->comment = current->comment;  /*new comments override old ones */
          }
        }
      }
      if( foundit == 0){  /* this is a new one */
         current->state = NEW;
         tail->next = current;
         tail = current;
      }
    }
  }
  strcpy(inline,"! end of NEW linker option file");
  characterize();
  tail->next=current;
  tail = current;
}

void writefinal(void){
ALINE *scanner;
char prefix[128];

 /* first pass, convert all UNKNOWN to REMOVED */

 for(scanner=head; scanner != NULL; scanner=scanner->next){
   if(scanner->state == UNKNOWN){
      switch (scanner->type) {
        case PSECT:
          scanner->type=PRIVATE_DATA;
          break;
        case PROCEDURE:
          scanner->type=PRIVATE_PROCEDURE;
          break;
      }
   }
 }

 /* second pass, write it out */

 for(scanner=head; scanner != NULL; scanner=scanner->next){
   switch (scanner->type) {
     case GSMATCH:
        fprintf(ffinal,"GSMATCH=lequal,%d,%d\n",major,minor);
        break;
     case PROCEDURE:
        fprintf(ffinal,"SYMBOL_VECTOR = (%s=PROCEDURE)%s\n",scanner->string,scanner->comment);
        break;
     case PSECT:
        fprintf(ffinal,"SYMBOL_VECTOR = (%s=PSECT)%s\n",scanner->string,scanner->comment);
        break;
     case PRIVATE_PROCEDURE:
        fprintf(ffinal,"SYMBOL_VECTOR = (MISSINGFCT=PRIVATE_PROCEDURE)%s\n",scanner->comment);
        break;
     case PRIVATE_DATA:
        fprintf(ffinal,"SYMBOL_VECTOR = (%s=PRIVATE_DATA)%s\n",scanner->string,scanner->comment);
        break;
     case OTHER:
        (void) strcpy(inline,scanner->string);
        (void) strcat(inline,scanner->comment);
        fprintf(ffinal,"%s\n",inline);
        break;
   }
 }
}



int main(int argc, char *argv[]){

  if(argc != 6){
    (void) fprintf(stderr,"upward_opt must be invoked with a command like\n\n");
    (void) fprintf(stderr,"upward_opt old.opt new.opt final.opt major minor \n\n");
    (void) fprintf(stderr,"  old.opt        linker option file for previous version\n");
    (void) fprintf(stderr,"  new.opt        linker option file for current version\n");
    (void) fprintf(stderr,"  final.opt      upwardly compatible linker option file for current version\n");
    (void) fprintf(stderr,"  major          Major version number, for GSMATCH\n");
    (void) fprintf(stderr,"  minor          Minor version number, for GSMATCH\n");
    exit(EXIT_FAILURE);
  }

  fold=fopen(argv[1],"r");
  if(fold == NULL){
    (void) fprintf(stderr,"upward_opt, fatal error, could not open file [%s]\n",argv[1]);
    exit(EXIT_FAILURE);
  }
  fnew=fopen(argv[2],"r");
  if(fnew == NULL){
    (void) fprintf(stderr,"upward_opt, fatal error, could not open file [%s]\n",argv[2]);
    exit(EXIT_FAILURE);
  }
  ffinal=fopen(argv[3],"w");
  if(ffinal == NULL){
    (void) fprintf(stderr,"upward_opt, fatal error, could not create file [%s]\n",argv[3]);
    exit(EXIT_FAILURE);
  }
  (void) sscanf(argv[4],"%d",&major);
  (void) sscanf(argv[5],"%d",&minor);

  /* read in the the OLD option file*/

  readold();

  /* read in the the NEW option file*/

  readnew();

  /* write out the FINAL option file*/

  writefinal();

  exit(EXIT_SUCCESS);
}
